//#define ALLOW_ACCESS_TO_INTERNALS_OF_MENUS
//#define ALLOW_ACCESS_TO_INTERNALS_OF_BITMAPS
#include <PalmOS.h>
#include <PceNativeCall.h>
#include <CtlGlue.h>

#define DA_STACK_SIZE ( 3 * 1024 )

#define APP_LIST_ID 3433

#define TEXT_OFFSET 18

#define appCreator 'meDA'
#define TYPE_HISTORY 'hist'

//register void *stackPointer asm("%sp") ;

void LaunchByTypeCreator( UInt32 type, UInt32 creator );

typedef struct {
    UInt32  type;
    UInt32  creator;
} PopupApp;


static void Choices( void );


void DAStart( void )
{
    EvtFlushKeyQueue();
    Choices();
}

static void GetName( Char* name, UInt32 type, UInt32 creatorID )
{
    DmOpenRef ref;
    MemHandle h;

    MemSet( name, dmDBNameLength, 0 );

    ref = DmOpenDatabaseByTypeCreator( type, creatorID, dmModeReadOnly );

    if ( ref == NULL )
        return;

    h = DmGet1Resource( 'tAIN', 1000 );

    if ( h != NULL ) {
        UInt16 len;
        Char*  p;

        p = MemHandleLock( h );

        len = MemHandleSize( h );

        if ( dmDBNameLength - 1 < len )
            len = dmDBNameLength - 1;

        MemMove( name, p, len );

        MemHandleUnlock( h );

        DmReleaseResource( h );
    }

    if ( *name == 0 ) {
        Err err;
        LocalID id;
        UInt16  card;

        err = DmOpenDatabaseInfo( ref, &id, NULL, NULL, &card, NULL );

        if ( err == errNone ) {
            DmDatabaseInfo( card, id, name, NULL, NULL,
                NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL );
        }
    }

    DmCloseDatabase( ref );
}


static Char** PopulateChoices( PopupApp* appList, UInt16* numEntriesP )
{
    UInt16 i;
    UInt16 out;
    Char** choices;

    choices = MemPtrNew( *numEntriesP * sizeof( Char* ) );

    if ( choices == NULL )
        return NULL;

    for ( out = 0, i = 0 ; i < *numEntriesP ; i++ ) {
        choices[ out ] = MemPtrNew( dmDBNameLength * sizeof( Char ) );
        if ( choices[ out ] == NULL ) {
            UInt16 j;

            for ( j = 0 ; j < out ; j++ )
                 MemPtrFree( choices[ out ] );

            MemPtrFree( choices );

            return NULL;
        }
        GetName( choices[ out ], appList[ i ].type, appList[ i ].creator );
        if ( choices[ out ][ 0 ] == 0 ) {
            MemPtrFree( choices[ out ] );
        }
        else {
            appList[ out ] = appList[ i ];
            out++;
        }
    }

    *numEntriesP = out;

    return choices;
}



static void CleanChoices( Char** choices, UInt16 numEntries )
{
    UInt16 i;

    for ( i = 0 ; i < numEntries ; i++ ) {
         MemPtrFree( choices[ i ] );
    }
    MemPtrFree( choices );

}



static Int16 AppPopup( PopupApp* appList, UInt16 numEntries )
{
    FormType* newForm;
    Coord     x;
    Coord     y;
    Coord     formHeight;
    Coord     formWidth;
    Int16     chosen;
    RectangleType formRect;
    Char**        choices;
    UInt16        i;
    RectangleType screen;
    Char          zero;
    Coord         extentX;
    Coord         extentY;

    zero = 0;

    if ( numEntries == 0 )
        return -1;

    chosen = -1;

    choices = PopulateChoices( appList, &numEntries );

    if ( choices == NULL )
        return -1;

    FntSetFont( stdFont );

    formHeight = numEntries * FntLineHeight();
    formWidth  = TEXT_OFFSET * 2;

    WinPushDrawState();

    for ( i = 0 ; i < numEntries ; i++ ) {
         Coord w;

         FntSetFont( stdFont );

         w = FntCharsWidth( choices[ i ], StrLen( choices[ i ] ) ) + TEXT_OFFSET + 5;

         if ( formWidth < w )
             formWidth = w;
    }

    WinPopDrawState();

    WinGetDisplayExtent( &( screen.extent.x ), &( screen.extent.y ) );

    x = 0;
    y = screen.extent.y - formHeight;

    formRect.topLeft.x = x - 4;
    formRect.topLeft.y = y - 4;
    formRect.extent.x  = formWidth + 8;
    formRect.extent.y  = formHeight + 8;

    newForm = FrmNewForm( FrmGetActiveFormID() + 1, &zero, x, y, formWidth, formHeight, false, 0, 0, 0 );

    if ( newForm != NULL ) {
        FormType* oldForm;
        UInt16    visibleEntries;

        oldForm = FrmGetActiveForm();

        FrmSetActiveForm( newForm );

        WinGetDisplayExtent( &extentX, &extentY );
        visibleEntries = extentY / FntLineHeight();

        if ( numEntries < visibleEntries )
            visibleEntries = numEntries;

        if ( errNone == LstNewList( ( void** )&newForm, APP_LIST_ID, 0, 0, formWidth,
                 formHeight, stdFont, visibleEntries, 0 ) ) {
            UInt16    listIndex;
            ListType* list;

            listIndex = FrmGetObjectIndex( newForm, APP_LIST_ID );

            list      = FrmGetObjectPtr( newForm, listIndex );

            LstSetListChoices( list, choices, numEntries );
            LstSetSelection( list, numEntries - 1 );

            chosen = LstPopupList( list );
        }

        FrmEraseForm( newForm );
        FrmDeleteForm( newForm );
        FrmSetActiveForm( oldForm );
    }

    CleanChoices( choices, numEntries );
    return chosen;
}



static UInt32 GetDate( DmOpenRef ref, UInt32 creator )
{
    UInt16 index;
    UInt32* dateP;
    UInt32  date;
    MemHandle h;

    if ( ref == NULL || 0xFFFF == ( index = DmFindResource( ref, creator, 0, NULL ) ) )
        return 0;

    h = DmGetResourceIndex( ref, index );
    
    date = 0;

    if ( MemHandleSize( h ) == sizeof( UInt32 ) ) {
        dateP = MemHandleLock( h );
        date  = *dateP;
        MemHandleUnlock( h );
    }

    return date;
}



static void SetDate( DmOpenRef ref, UInt32 creator, UInt32 date )
{
    UInt16 index;
    MemHandle h;
    Boolean   release;
    
    release = false;

    index = DmFindResource( ref, creator, 0, NULL );

    if ( index == 0xFFFF ) {
        h = DmNewResource( ref, creator, 0, sizeof( UInt32 ) );
        release = true;
    }
    else {
        h = DmGetResourceIndex( ref, index );
        if ( MemHandleSize( h ) != sizeof( UInt32 ) ) {
            h = DmResizeResource( h, sizeof( UInt32 ) );
            release = true;
        }
    }

    if ( h != NULL ) {
        UInt32* p;

        p = MemHandleLock( h );

        DmWrite( p, 0, &date, sizeof( UInt32 ) );

        MemHandleUnlock( h );

        if ( release ) {
            DmReleaseResource( h );
        }
    }
}



static Int16 Compare( void* a, void* b, Int32 other )
{
    PopupApp* appA;
    PopupApp* appB;
    Char      nameA[ dmDBNameLength ];
    Char      nameB[ dmDBNameLength ];
    UInt32    dateA;
    UInt32    dateB;

    appA = ( PopupApp* )a;
    appB = ( PopupApp* )b;

    dateA = GetDate( ( DmOpenRef )other, appA->creator );
    dateB = GetDate( ( DmOpenRef )other, appB->creator );

    if ( dateA < dateB ) {
        return -1;
    }
    else {
        return 1;
    }

    GetName( nameA, appA->type, appA->creator );
    GetName( nameB, appB->type, appB->creator );

    return StrCaselessCompare( nameA, nameB );
}



static void CreateHistory( void )
{
    Char name[ dmDBNameLength ];
    Char* p;
    
    p = name;

    *p++ = 'm';
    *p++ = 'e';
    *p++ = 'D';
    *p++ = 'A';
    *p++ = '-';
    *p++ = 'H';
    *p++ = 'i';
    *p++ = 's';
    *p++ = 't';
    *p++ = 'o';
    *p++ = 'r';
    *p++ = 'y';
    *p++ = 0;

    DmCreateDatabase( 0, name, appCreator, TYPE_HISTORY, true );
}



static void Choices( void )
{
    PopupApp* appList;
    Int16         chosen;
    UInt16        count;
    UInt16        size;
    UInt16        i;
    DmSearchStateType state;
    Boolean first;
    UInt16 card;
    LocalID id;

    first = true;

    count = 0;

    while ( errNone == DmGetNextDatabaseByTypeCreator( first, &state, 'DAcc', 0,
                             false, &card, &id ) ) {
        count++;
        first = false;
    }

    if ( count <= 1 )
        return;

    appList = MemPtrNew( count * sizeof( PopupApp ) );
    if ( appList == NULL )
        return;

    i = 0;

    first = true;

    while ( errNone == DmGetNextDatabaseByTypeCreator( first, &state, 'DAcc', 0,
                             false, &card, &id ) && i < count ) {
        if ( errNone == DmDatabaseInfo( card, id, NULL, NULL, NULL, NULL, NULL, NULL,
                                    NULL, NULL, NULL, &( appList[ i ].type ),
                                      &( appList[ i ].creator ) ) && appList[ i ].creator != appCreator ) {
            i++;
        }

        first = false;
    }

    count  = i;

    if ( 0 < count ) {
        DmOpenRef dateRef;

        CreateHistory();

        dateRef = DmOpenDatabaseByTypeCreator( TYPE_HISTORY, appCreator, dmModeReadWrite );

        SysInsertionSort( appList, count, sizeof( *appList ), Compare, ( Int32 )dateRef );

        chosen = AppPopup( appList, count );

        if ( 0 <= chosen ) {
            if ( dateRef != NULL ) {
                SetDate( dateRef, appList[ chosen ].creator, TimGetSeconds() );
                DmCloseDatabase( dateRef );
            }

            LaunchByTypeCreator( appList[ chosen ].type, appList[ chosen ].creator );
        }
        else {
            if ( dateRef != NULL ) {
                DmCloseDatabase( dateRef );
            }
        }
    }

    MemPtrFree( appList );
}



static void LaunchDA( LocalID id )
{
    DmOpenRef ref;
    MemHandle h;

    void (*f)( void );

    ref = DmOpenDatabase( 0, id, dmModeReadOnly );

    if ( ref == NULL )
        return;

    h = DmGet1Resource( 'code', 1000 );

    if ( h != NULL ) {
        WinHandle oldDrawWin;
        FormType* oldForm;
        WinHandle oldActiveWin;
//        UInt8*    stack;
//        UInt8*    oldStack;

        oldDrawWin = WinGetDrawWindow();
        oldActiveWin = WinGetActiveWindow();
        oldForm      = FrmGetActiveForm();
        f = MemHandleLock( h );

#if 0
        stack = MemPtrNew( DA_STACK_SIZE );

        oldStack = stackPointer;

        if ( stack != NULL ) {
            stackPointer = stack + DA_STACK_SIZE;
        }
#endif

        f();

#if 0
        if ( stack != NULL ) {
            stackPointer = oldStack;
        }
#endif

        MemHandleUnlock( h );
        DmReleaseResource( h );
        if ( oldForm != NULL )
            FrmSetActiveForm( oldForm );
        if ( oldDrawWin != NULL )
            WinSetDrawWindow( oldDrawWin );
        if ( oldActiveWin != NULL )
            WinSetActiveWindow( oldActiveWin );
    }
    DmCloseDatabase( ref );
}



void LaunchByTypeCreatorID( UInt32 type, UInt32 creator, LocalID id )
{
    if ( creator == 'lnch' && type == 'appl' ) {
        EvtEnqueueKey( vchrLaunch, 0, commandKeyMask );
        return;
    }

    if ( type == 'DAcc' )
        LaunchDA( id );
    else if ( type == 'panl' )
        SysUIAppSwitch( 0, id, sysAppLaunchCmdPanelCalledFromApp, NULL );
    else
        SysUIAppSwitch( 0, id, 0, NULL );
}



void LaunchByTypeCreator( UInt32 type, UInt32 creator )
{
    DmSearchStateType state;
    Err err;
    UInt16 card;
    LocalID id;

    err = DmGetNextDatabaseByTypeCreator( true, &state, type, creator, true, &card, &id );

    if ( err == errNone )
        LaunchByTypeCreatorID( type, creator, id );
}



